# Copyright (C) 2012-2025 Ben Kurtovic <ben.kurtovic@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from ..nodes import (
    Argument,
    Comment,
    ExternalLink,
    Heading,
    HTMLEntity,
    Tag,
    Template,
    Text,
    Wikilink,
)
from ..nodes.extras import Attribute, Parameter
from ..smart_list import SmartList
from ..wikicode import Wikicode
from . import tokens
from .errors import ParserError

__all__ = ["Builder"]

_HANDLERS = {tokens.Text: lambda self, token: Text(token.text)}


def _add_handler(token_type):
    """Create a decorator that adds a handler function to the lookup table."""

    def decorator(func):
        """Add a handler function to the lookup table."""
        _HANDLERS[token_type] = func
        return func

    return decorator


class Builder:
    """Builds a tree of nodes out of a sequence of tokens.

    To use, pass a list of :class:`.Token`\\ s to the :meth:`build` method. The
    list will be exhausted as it is parsed and a :class:`.Wikicode` object
    containing the node tree will be returned.
    """

    def __init__(self):
        self._tokens = []
        self._stacks = []

    def _push(self):
        """Push a new node list onto the stack."""
        self._stacks.append([])

    def _pop(self):
        """Pop the current node list off of the stack.

        The raw node list is wrapped in a :class:`.SmartList` and then in a
        :class:`.Wikicode` object.
        """
        return Wikicode(SmartList(self._stacks.pop()))

    def _write(self, item):
        """Append a node to the current node list."""
        self._stacks[-1].append(item)

    def _handle_parameter(self, default):
        """Handle a case where a parameter is at the head of the tokens.

        *default* is the value to use if no parameter name is defined.
        """
        key = None
        showkey = False
        self._push()
        while self._tokens:
            token = self._tokens.pop()
            if isinstance(token, tokens.TemplateParamEquals):
                key = self._pop()
                showkey = True
                self._push()
            elif isinstance(
                token, (tokens.TemplateParamSeparator, tokens.TemplateClose)
            ):
                self._tokens.append(token)
                value = self._pop()
                if key is None:
                    key = Wikicode(SmartList([Text(str(default))]))
                return Parameter(key, value, showkey)
            else:
                self._write(self._handle_token(token))
        raise ParserError("_handle_parameter() missed a close token")

    @_add_handler(tokens.TemplateOpen)
    def _handle_template(self, token):
        """Handle a case where a template is at the head of the tokens."""
        name = None
        params = []
        default = 1
        self._push()
        while self._tokens:
            token = self._tokens.pop()
            if isinstance(token, tokens.TemplateParamSeparator):
                if not params:
                    name = self._pop()
                param = self._handle_parameter(default)
                params.append(param)
                if not param.showkey:
                    default += 1
            elif isinstance(token, tokens.TemplateClose):
                if not params:
                    name = self._pop()
                assert name is not None
                return Template(name, params)
            else:
                self._write(self._handle_token(token))
        raise ParserError("_handle_template() missed a close token")

    @_add_handler(tokens.ArgumentOpen)
    def _handle_argument(self, token):
        """Handle a case where an argument is at the head of the tokens."""
        name = None
        self._push()
        while self._tokens:
            token = self._tokens.pop()
            if isinstance(token, tokens.ArgumentSeparator):
                name = self._pop()
                self._push()
            elif isinstance(token, tokens.ArgumentClose):
                if name is not None:
                    return Argument(name, self._pop())
                return Argument(self._pop())
            else:
                self._write(self._handle_token(token))
        raise ParserError("_handle_argument() missed a close token")

    @_add_handler(tokens.WikilinkOpen)
    def _handle_wikilink(self, token):
        """Handle a case where a wikilink is at the head of the tokens."""
        title = None
        self._push()
        while self._tokens:
            token = self._tokens.pop()
            if isinstance(token, tokens.WikilinkSeparator):
                title = self._pop()
                self._push()
            elif isinstance(token, tokens.WikilinkClose):
                if title is not None:
                    return Wikilink(title, self._pop())
                return Wikilink(self._pop())
            else:
                self._write(self._handle_token(token))
        raise ParserError("_handle_wikilink() missed a close token")

    @_add_handler(tokens.ExternalLinkOpen)
    def _handle_external_link(self, token):
        """Handle when an external link is at the head of the tokens."""
        brackets, url, suppress_space = token.brackets, None, None
        self._push()
        while self._tokens:
            token = self._tokens.pop()
            if isinstance(token, tokens.ExternalLinkSeparator):
                url = self._pop()
                suppress_space = token.suppress_space
                self._push()
            elif isinstance(token, tokens.ExternalLinkClose):
                if url is not None:
                    return ExternalLink(
                        url,
                        self._pop(),
                        brackets=brackets,
                        suppress_space=suppress_space is True,
                    )
                return ExternalLink(
                    self._pop(),
                    brackets=brackets,
                    suppress_space=suppress_space is True,
                )
            else:
                self._write(self._handle_token(token))
        raise ParserError("_handle_external_link() missed a close token")

    @_add_handler(tokens.HTMLEntityStart)
    def _handle_entity(self, token):
        """Handle a case where an HTML entity is at the head of the tokens."""
        token = self._tokens.pop()
        if isinstance(token, tokens.HTMLEntityNumeric):
            token = self._tokens.pop()
            if isinstance(token, tokens.HTMLEntityHex):
                text = self._tokens.pop()
                self._tokens.pop()  # Remove HTMLEntityEnd
                return HTMLEntity(
                    text.text, named=False, hexadecimal=True, hex_char=token.char
                )
            self._tokens.pop()  # Remove HTMLEntityEnd
            return HTMLEntity(token.text, named=False, hexadecimal=False)
        self._tokens.pop()  # Remove HTMLEntityEnd
        return HTMLEntity(token.text, named=True, hexadecimal=False)

    @_add_handler(tokens.HeadingStart)
    def _handle_heading(self, token):
        """Handle a case where a heading is at the head of the tokens."""
        level = token.level
        self._push()
        while self._tokens:
            token = self._tokens.pop()
            if isinstance(token, tokens.HeadingEnd):
                title = self._pop()
                return Heading(title, level)
            self._write(self._handle_token(token))
        raise ParserError("_handle_heading() missed a close token")

    @_add_handler(tokens.CommentStart)
    def _handle_comment(self, token):
        """Handle a case where an HTML comment is at the head of the tokens."""
        self._push()
        while self._tokens:
            token = self._tokens.pop()
            if isinstance(token, tokens.CommentEnd):
                contents = self._pop()
                return Comment(contents)
            self._write(self._handle_token(token))
        raise ParserError("_handle_comment() missed a close token")

    def _handle_attribute(self, start):
        """Handle a case where a tag attribute is at the head of the tokens."""
        name = quotes = None
        self._push()
        while self._tokens:
            token = self._tokens.pop()
            if isinstance(token, tokens.TagAttrEquals):
                name = self._pop()
                self._push()
            elif isinstance(token, tokens.TagAttrQuote):
                quotes = token.char
            elif isinstance(
                token,
                (tokens.TagAttrStart, tokens.TagCloseOpen, tokens.TagCloseSelfclose),
            ):
                self._tokens.append(token)
                if name:
                    value = self._pop()
                else:
                    name, value = self._pop(), None
                return Attribute(
                    name,
                    value,
                    quotes,
                    start.pad_first,
                    start.pad_before_eq,
                    start.pad_after_eq,
                )
            else:
                self._write(self._handle_token(token))
        raise ParserError("_handle_attribute() missed a close token")

    @_add_handler(tokens.TagOpenOpen)
    def _handle_tag(self, token):
        """Handle a case where a tag is at the head of the tokens."""
        close_tokens = (tokens.TagCloseSelfclose, tokens.TagCloseClose)
        tag = None
        padding = None
        implicit, attrs, contents, closing_tag = False, [], None, None
        wiki_markup, invalid = token.wiki_markup, token.invalid or False
        wiki_style_separator, closing_wiki_markup = None, wiki_markup
        self._push()
        while self._tokens:
            token = self._tokens.pop()
            if isinstance(token, tokens.TagAttrStart):
                attrs.append(self._handle_attribute(token))
            elif isinstance(token, tokens.TagCloseOpen):
                wiki_style_separator = token.wiki_markup
                padding = token.padding or ""
                tag = self._pop()
                self._push()
            elif isinstance(token, tokens.TagOpenClose):
                closing_wiki_markup = token.wiki_markup
                contents = self._pop()
                self._push()
            elif isinstance(token, close_tokens):
                if isinstance(token, tokens.TagCloseSelfclose):
                    closing_wiki_markup = token.wiki_markup
                    tag = self._pop()
                    self_closing = True
                    padding = token.padding or ""
                    implicit = token.implicit or False
                else:
                    self_closing = False
                    closing_tag = self._pop()
                assert tag is not None
                assert padding is not None
                return Tag(
                    tag,
                    contents,
                    attrs,
                    wiki_markup,
                    self_closing,
                    invalid,
                    implicit,
                    padding,
                    closing_tag,
                    wiki_style_separator,
                    closing_wiki_markup,
                )
            else:
                self._write(self._handle_token(token))
        raise ParserError("_handle_tag() missed a close token")

    def _handle_token(self, token):
        """Handle a single token."""
        try:
            return _HANDLERS[type(token)](self, token)
        except KeyError:
            err = "_handle_token() got unexpected {0}"
            raise ParserError(err.format(type(token).__name__)) from None

    def build(self, tokenlist):
        """Build a Wikicode object from a list tokens and return it."""
        self._tokens = tokenlist
        self._tokens.reverse()
        self._push()
        while self._tokens:
            node = self._handle_token(self._tokens.pop())
            self._write(node)
        return self._pop()


del _add_handler



# Copyright (C) 2012-2025 Ben Kurtovic <ben.kurtovic@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

__all__ = ["ParserError"]


class ParserError(Exception):
    """Exception raised when an internal error occurs while parsing.

    This does not mean that the wikicode was invalid, because invalid markup
    should still be parsed correctly. This means that the parser caught itself
    with an impossible internal state and is bailing out before other problems
    can happen. Its appearance indicates a bug.
    """

    def __init__(self, extra):
        msg = f"This is a bug and should be reported. Info: {extra}."
        super().__init__(msg)



# Copyright (C) 2012-2025 Ben Kurtovic <ben.kurtovic@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from ...string_mixin import StringMixIn
from ...utils import parse_anything

if TYPE_CHECKING:
    from ...wikicode import Wikicode

__all__ = ["Attribute"]


class Attribute(StringMixIn):
    """Represents an attribute of an HTML tag.

    This is used by :class:`.Tag` objects. For example, the tag
    ``<ref name="foo">`` contains an Attribute whose name is ``"name"`` and
    whose value is ``"foo"``.
    """

    def __init__(
        self,
        name: Any,
        value: Any = None,
        quotes: str | None = '"',
        pad_first: str = " ",
        pad_before_eq: str = "",
        pad_after_eq: str = "",
    ):
        super().__init__()

        self._pad_first: str
        self._pad_before_eq: str
        self._pad_after_eq: str

        self.name = name
        self._quotes: str | None = None
        self.value = value
        self.quotes = quotes
        self.pad_first = pad_first
        self.pad_before_eq = pad_before_eq
        self.pad_after_eq = pad_after_eq

    def __str__(self) -> str:
        result = self.pad_first + str(self.name) + self.pad_before_eq
        if self.value is not None:
            result += "=" + self.pad_after_eq
            if self.quotes:
                return result + self.quotes + str(self.value) + self.quotes
            return result + str(self.value)
        return result

    @staticmethod
    def _value_needs_quotes(value: Wikicode | None) -> str | None:
        """Return valid quotes for the given value, or None if unneeded."""
        if not value:
            return None
        val = "".join(str(node) for node in value.filter_text(recursive=False))
        if not any(char.isspace() for char in val):
            return None
        if "'" in val and '"' not in val:
            return '"'
        if '"' in val and "'" not in val:
            return "'"
        return "\"'"  # Either acceptable, " preferred over '

    def _set_padding(self, attr: str, value: str) -> None:
        """Setter for the value of a padding attribute."""
        if not value:
            setattr(self, attr, "")
        else:
            value = str(value)
            if not value.isspace():
                raise ValueError("padding must be entirely whitespace")
            setattr(self, attr, value)

    @staticmethod
    def coerce_quotes(quotes: Any) -> str | None:
        """Coerce a quote type into an acceptable value, or raise an error."""
        coerced_quotes = str(quotes) if quotes else None
        if coerced_quotes not in [None, '"', "'"]:
            raise ValueError(f"{quotes!r} is not a valid quote type")
        return coerced_quotes

    @property
    def name(self) -> Wikicode:
        """The name of the attribute as a :class:`.Wikicode` object."""
        return self._name

    @name.setter
    def name(self, value: Any) -> None:
        self._name = parse_anything(value)

    @property
    def value(self) -> Wikicode | None:
        """The value of the attribute as a :class:`.Wikicode` object."""
        return self._value

    @value.setter
    def value(self, newval: Any) -> None:
        if newval is None:
            self._value = None
        else:
            code = parse_anything(newval)
            quotes = self._value_needs_quotes(code)
            if quotes and (not self.quotes or self.quotes not in quotes):
                self._quotes = quotes[0]
            self._value = code

    @property
    def quotes(self) -> str | None:
        """How to enclose the attribute value. ``"``, ``'``, or ``None``."""
        return self._quotes

    @quotes.setter
    def quotes(self, value: Any) -> None:
        value = self.coerce_quotes(value)
        if not value and self._value_needs_quotes(self.value):
            raise ValueError("attribute value requires quotes")
        self._quotes = value

    @property
    def pad_first(self) -> str:
        """Spacing to insert right before the attribute."""
        return self._pad_first

    @pad_first.setter
    def pad_first(self, value: str) -> None:
        self._set_padding("_pad_first", value)

    @property
    def pad_before_eq(self) -> str:
        """Spacing to insert right before the equal sign."""
        return self._pad_before_eq

    @pad_before_eq.setter
    def pad_before_eq(self, value: str) -> None:
        self._set_padding("_pad_before_eq", value)

    @property
    def pad_after_eq(self) -> str:
        """Spacing to insert right after the equal sign."""
        return self._pad_after_eq

    @pad_after_eq.setter
    def pad_after_eq(self, value: str) -> None:
        self._set_padding("_pad_after_eq", value)



# Copyright (C) 2012-2025 Ben Kurtovic <ben.kurtovic@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Contains data about certain markup, like HTML tags and external links.

When updating this file, please also update the the C tokenizer version:
- mwparserfromhell/parser/ctokenizer/definitions.c
- mwparserfromhell/parser/ctokenizer/definitions.h
"""

from __future__ import annotations

__all__ = [
    "get_html_tag",
    "is_parsable",
    "is_visible",
    "is_single",
    "is_single_only",
    "is_scheme",
]

URI_SCHEMES = {
    # [wikimedia/mediawiki.git]/includes/DefaultSettings.php @ 5c660de5d0
    "bitcoin": False,
    "ftp": True,
    "ftps": True,
    "geo": False,
    "git": True,
    "gopher": True,
    "http": True,
    "https": True,
    "irc": True,
    "ircs": True,
    "magnet": False,
    "mailto": False,
    "mms": True,
    "news": False,
    "nntp": True,
    "redis": True,
    "sftp": True,
    "sip": False,
    "sips": False,
    "sms": False,
    "ssh": True,
    "svn": True,
    "tel": False,
    "telnet": True,
    "urn": False,
    "worldwind": True,
    "xmpp": False,
}

PARSER_BLACKLIST = [
    # https://www.mediawiki.org/wiki/Parser_extension_tags @ 2020-12-21
    "categorytree",
    "ce",
    "chem",
    "gallery",
    "graph",
    "hiero",
    "imagemap",
    "inputbox",
    "math",
    "nowiki",
    "pre",
    "score",
    "section",
    "source",
    "syntaxhighlight",
    "templatedata",
    "timeline",
]

INVISIBLE_TAGS = [
    # https://www.mediawiki.org/wiki/Parser_extension_tags @ 2020-12-21
    "categorytree",
    "gallery",
    "graph",
    "imagemap",
    "inputbox",
    "math",
    "score",
    "section",
    "templatedata",
    "timeline",
]

# [wikimedia/mediawiki.git]/includes/parser/Sanitizer.php @ 95e17ee645
SINGLE_ONLY = ["br", "wbr", "hr", "meta", "link", "img"]
SINGLE = SINGLE_ONLY + ["li", "dt", "dd", "th", "td", "tr"]

MARKUP_TO_HTML = {
    "#": "li",
    "*": "li",
    ";": "dt",
    ":": "dd",
}


def get_html_tag(markup: str) -> str:
    """Return the HTML tag associated with the given wiki-markup."""
    return MARKUP_TO_HTML[markup]


def is_parsable(tag: str) -> bool:
    """Return if the given *tag*'s contents should be passed to the parser."""
    return tag.lower() not in PARSER_BLACKLIST


def is_visible(tag: str) -> bool:
    """Return whether or not the given *tag* contains visible text."""
    return tag.lower() not in INVISIBLE_TAGS


def is_single(tag: str) -> bool:
    """Return whether or not the given *tag* can exist without a close tag."""
    return tag.lower() in SINGLE


def is_single_only(tag: str) -> bool:
    """Return whether or not the given *tag* must exist without a close tag."""
    return tag.lower() in SINGLE_ONLY


def is_scheme(scheme: str, slashes: bool = True) -> bool:
    """Return whether *scheme* is valid for external links."""
    scheme = scheme.lower()
    if slashes:
        return scheme in URI_SCHEMES
    return scheme in URI_SCHEMES and not URI_SCHEMES[scheme]



# Copyright (C) 2012-2025 Ben Kurtovic <ben.kurtovic@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Generator
from typing import TYPE_CHECKING, Any, Callable

from ..string_mixin import StringMixIn

if TYPE_CHECKING:
    from ..wikicode import Wikicode

__all__ = ["Node"]


class Node(StringMixIn):
    """Represents the base Node type, demonstrating the methods to override.

    :meth:`__str__` must be overridden. It should return a ``str``
    representation of the node. If the node contains :class:`.Wikicode`
    objects inside of it, :meth:`__children__` should be a generator that
    iterates over them. If the node is printable (shown when the page is
    rendered), :meth:`__strip__` should return its printable version,
    stripping out any formatting marks. It does not have to return a string,
    but something that can be converted to a string with ``str()``. Finally,
    :meth:`__showtree__` can be overridden to build a nice tree representation
    of the node, if desired, for :meth:`~.Wikicode.get_tree`.
    """

    def __str__(self) -> str:
        raise NotImplementedError()

    def __children__(self) -> Generator[Wikicode, None, None]:
        return
        # pylint: disable=unreachable
        yield  # pragma: no cover (this is a generator that yields nothing)

    def __strip__(self, **kwargs: Any) -> str | None:
        return None

    def __showtree__(
        self,
        write: Callable[[str], None],
        get: Callable[[Wikicode], None],
        mark: Callable[[], None],
    ) -> None:
        write(str(self))



# Copyright (C) 2012-2025 Ben Kurtovic <ben.kurtovic@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
This module contains the :class:`.StringMixIn` type, which implements the
interface for the ``str`` type in a dynamic manner.
"""

from __future__ import annotations

import sys
from collections.abc import Iterable, Iterator, Mapping
from typing import Any, SupportsIndex

__all__ = ["StringMixIn"]


def inheritdoc(method):
    """Set __doc__ of *method* to __doc__ of *method* in its parent class.

    Since this is used on :class:`.StringMixIn`, the "parent class" used is
    ``str``. This function can be used as a decorator.
    """
    method.__doc__ = getattr(str, method.__name__).__doc__
    return method


class StringMixIn:
    """Implement the interface for ``str`` in a dynamic manner.

    To use this class, inherit from it and override the :meth:`__str__` method
    to return the string representation of the object. The various string
    methods will operate on the value of :meth:`__str__` instead of the
    immutable ``self`` like the regular ``str`` type.
    """

    # This is based on collections.UserString, but:
    # - Requires overriding __str__ instead of setting .data
    # - Returns new strings as strs instead of StringMixIns

    __slots__ = ()

    def __str__(self) -> str:
        raise NotImplementedError()

    def __bytes__(self) -> bytes:
        return bytes(str(self), sys.getdefaultencoding())

    def __repr__(self) -> str:
        return repr(str(self))

    def __lt__(self, other: str | StringMixIn) -> bool:
        return str(self) < other

    def __le__(self, other: str | StringMixIn) -> bool:
        return str(self) <= other

    def __eq__(self, other: Any) -> bool:
        return str(self) == other

    def __ne__(self, other: Any) -> bool:
        return str(self) != other

    def __gt__(self, other: str | StringMixIn) -> bool:
        return str(self) > other

    def __ge__(self, other: str | StringMixIn) -> bool:
        return str(self) >= other

    def __bool__(self) -> bool:
        return bool(str(self))

    def __len__(self) -> int:
        return len(str(self))

    def __iter__(self) -> Iterator[str]:
        yield from str(self)

    def __getitem__(self, key: SupportsIndex | slice) -> str:
        return str(self)[key]

    def __reversed__(self) -> Iterator[str]:
        return reversed(str(self))

    def __contains__(self, item: Any) -> bool:
        return str(item) in str(self)

    @inheritdoc
    def capitalize(self) -> str:
        return str(self).capitalize()

    @inheritdoc
    def casefold(self) -> str:
        return str(self).casefold()

    @inheritdoc
    def center(self, width: int, fillchar: str = " ") -> str:
        return str(self).center(width, fillchar)

    @inheritdoc
    def count(
        self,
        sub: str | StringMixIn,
        start: SupportsIndex | None = None,
        end: SupportsIndex | None = None,
    ) -> int:
        if isinstance(sub, StringMixIn):
            sub = str(sub)
        return str(self).count(sub, start, end)

    @inheritdoc
    def encode(self, encoding: str = "utf-8", errors: str = "strict") -> bytes:
        return str(self).encode(encoding, errors)

    @inheritdoc
    def endswith(
        self,
        suffix: str | tuple[str, ...],
        start: SupportsIndex | None = None,
        end: SupportsIndex | None = None,
    ) -> bool:
        return str(self).endswith(suffix, start, end)

    @inheritdoc
    def expandtabs(self, tabsize: int = 8) -> str:
        return str(self).expandtabs(tabsize)

    @inheritdoc
    def find(
        self,
        sub: str | StringMixIn,
        start: SupportsIndex | None = None,
        end: SupportsIndex | None = None,
    ) -> int:
        if isinstance(sub, StringMixIn):
            sub = str(sub)
        return str(self).find(sub, start, end)

    @inheritdoc
    def format(self, /, *args: Any, **kwds: Any) -> str:
        return str(self).format(*args, **kwds)

    @inheritdoc
    def format_map(self, mapping: Mapping[str, Any]) -> str:
        return str(self).format_map(mapping)

    @inheritdoc
    def index(
        self,
        sub: str,
        start: SupportsIndex | None = None,
        end: SupportsIndex | None = None,
    ) -> int:
        return str(self).index(sub, start, end)

    @inheritdoc
    def isalpha(self) -> bool:
        return str(self).isalpha()

    @inheritdoc
    def isalnum(self) -> bool:
        return str(self).isalnum()

    @inheritdoc
    def isascii(self) -> bool:
        return str(self).isascii()

    @inheritdoc
    def isdecimal(self) -> bool:
        return str(self).isdecimal()

    @inheritdoc
    def isdigit(self) -> bool:
        return str(self).isdigit()

    @inheritdoc
    def isidentifier(self) -> bool:
        return str(self).isidentifier()

    @inheritdoc
    def islower(self) -> bool:
        return str(self).islower()

    @inheritdoc
    def isnumeric(self) -> bool:
        return str(self).isnumeric()

    @inheritdoc
    def isprintable(self) -> bool:
        return str(self).isprintable()

    @inheritdoc
    def isspace(self) -> bool:
        return str(self).isspace()

    @inheritdoc
    def istitle(self) -> bool:
        return str(self).istitle()

    @inheritdoc
    def isupper(self) -> bool:
        return str(self).isupper()

    @inheritdoc
    def join(self, seq: Iterable[str]) -> str:
        return str(self).join(seq)

    @inheritdoc
    def ljust(self, width: int, fillchar: str = " ") -> str:
        return str(self).ljust(width, fillchar)

    @inheritdoc
    def lower(self) -> str:
        return str(self).lower()

    @inheritdoc
    def lstrip(self, chars: str | None = None) -> str:
        return str(self).lstrip(chars)

    maketrans = str.maketrans

    @inheritdoc
    def partition(self, sep: str) -> tuple[str, str, str]:
        return str(self).partition(sep)

    @inheritdoc
    def removeprefix(self, prefix: str | StringMixIn, /) -> str:
        if isinstance(prefix, StringMixIn):
            prefix = str(prefix)
        return str(self).removeprefix(prefix)

    @inheritdoc
    def removesuffix(self, suffix: str | StringMixIn, /) -> str:
        if isinstance(suffix, StringMixIn):
            suffix = str(suffix)
        return str(self).removesuffix(suffix)

    @inheritdoc
    def replace(
        self,
        old: str | StringMixIn,
        new: str | StringMixIn,
        /,
        count: SupportsIndex = -1,
    ):
        if isinstance(old, StringMixIn):
            old = str(old)
        if isinstance(new, StringMixIn):
            new = str(new)
        return str(self).replace(old, new, count)

    @inheritdoc
    def rfind(
        self,
        sub: str | StringMixIn,
        start: SupportsIndex | None = None,
        end: SupportsIndex | None = None,
    ) -> int:
        if isinstance(sub, StringMixIn):
            sub = str(sub)
        return str(self).rfind(sub, start, end)

    @inheritdoc
    def rindex(
        self,
        sub: str,
        start: SupportsIndex | None = None,
        end: SupportsIndex | None = None,
    ) -> int:
        return str(self).rindex(sub, start, end)

    @inheritdoc
    def rjust(self, width: int, fillchar: str = " ") -> str:
        return str(self).rjust(width, fillchar)

    @inheritdoc
    def rpartition(self, sep: str) -> tuple[str, str, str]:
        return str(self).rpartition(sep)

    @inheritdoc
    def rstrip(self, chars: str | None = None) -> str:
        return str(self).rstrip(chars)

    @inheritdoc
    def split(self, sep: str | None = None, maxsplit: SupportsIndex = -1) -> list[str]:
        return str(self).split(sep, maxsplit)

    @inheritdoc
    def rsplit(self, sep: str | None = None, maxsplit: SupportsIndex = -1) -> list[str]:
        return str(self).rsplit(sep, maxsplit)

    @inheritdoc
    def splitlines(self, keepends: bool = False) -> list[str]:
        return str(self).splitlines(keepends)

    @inheritdoc
    def startswith(
        self,
        prefix: str | tuple[str, ...],
        start: SupportsIndex | None = None,
        end: SupportsIndex | None = None,
    ) -> bool:
        return str(self).startswith(prefix, start, end)

    @inheritdoc
    def strip(self, chars: str | None = None) -> str:
        return str(self).strip(chars)

    @inheritdoc
    def swapcase(self) -> str:
        return str(self).swapcase()

    @inheritdoc
    def title(self) -> str:
        return str(self).title()

    @inheritdoc
    def translate(self, *args: Any) -> str:
        return str(self).translate(*args)

    @inheritdoc
    def upper(self) -> str:
        return str(self).upper()

    @inheritdoc
    def zfill(self, width: int) -> str:
        return str(self).zfill(width)



# Copyright (C) 2012-2025 Ben Kurtovic <ben.kurtovic@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Generator
from typing import TYPE_CHECKING, Any, Callable

from ..definitions import is_visible
from ..utils import parse_anything
from ._base import Node
from .extras import Attribute

if TYPE_CHECKING:
    from ..wikicode import Wikicode

__all__ = ["Tag"]


class Tag(Node):
    """Represents an HTML-style tag in wikicode, like ``<ref>``."""

    def __init__(
        self,
        tag: Any,
        contents: Any = None,
        attrs: list[Attribute] | None = None,
        wiki_markup: str | None = None,
        self_closing: bool = False,
        invalid: bool = False,
        implicit: bool = False,
        padding: str = "",
        closing_tag: Wikicode | None = None,
        wiki_style_separator: str | None = None,
        closing_wiki_markup: str | None = None,
    ):
        super().__init__()
        self._attrs: list[Attribute]
        self._closing_wiki_markup: str | None

        self.tag = tag
        self.contents = contents
        self._attrs = attrs if attrs else []
        self._closing_wiki_markup = None
        self.wiki_markup = wiki_markup
        self.self_closing = self_closing
        self.invalid = invalid
        self.implicit = implicit
        self.padding = padding
        if closing_tag is not None:
            self.closing_tag = closing_tag
        self.wiki_style_separator = wiki_style_separator
        if closing_wiki_markup is not None:
            self.closing_wiki_markup = closing_wiki_markup

    def __str__(self) -> str:
        if self.wiki_markup:
            if self.attributes:
                attrs = "".join([str(attr) for attr in self.attributes])
            else:
                attrs = ""
            padding = self.padding or ""
            separator = self.wiki_style_separator or ""
            if self.self_closing:
                return self.wiki_markup + attrs + padding + separator
            close = self.closing_wiki_markup or ""
            return (
                self.wiki_markup
                + attrs
                + padding
                + separator
                + str(self.contents)
                + close
            )

        result = ("</" if self.invalid else "<") + str(self.tag)
        if self.attributes:
            result += "".join([str(attr) for attr in self.attributes])
        if self.self_closing:
            result += self.padding + (">" if self.implicit else "/>")
        else:
            result += self.padding + ">" + str(self.contents)
            result += "</" + str(self.closing_tag) + ">"
        return result

    def __children__(self) -> Generator[Wikicode, None, None]:
        if not self.wiki_markup:
            yield self.tag
        for attr in self.attributes:
            yield attr.name
            if attr.value is not None:
                yield attr.value
        if not self.self_closing:
            yield self.contents
            if not self.wiki_markup and self.closing_tag:
                yield self.closing_tag

    def __strip__(self, **kwargs: Any) -> str | None:
        if self.contents and is_visible(str(self.tag)):
            return self.contents.strip_code(**kwargs)
        return None

    def __showtree__(
        self,
        write: Callable[[str], None],
        get: Callable[[Wikicode], None],
        mark: Callable[[], None],
    ) -> None:
        write("</" if self.invalid else "<")
        get(self.tag)
        for attr in self.attributes:
            get(attr.name)
            if not attr.value:
                continue
            write("    = ")
            mark()
            get(attr.value)
        if self.self_closing:
            write(">" if self.implicit else "/>")
        else:
            write(">")
            get(self.contents)
            write("</")
            get(self.closing_tag)
            write(">")

    @property
    def tag(self) -> Wikicode:
        """The tag itself, as a :class:`.Wikicode` object."""
        return self._tag

    @tag.setter
    def tag(self, value: Any) -> None:
        self._tag = self._closing_tag = parse_anything(value)

    @property
    def contents(self) -> Wikicode:
        """The contents of the tag, as a :class:`.Wikicode` object."""
        return self._contents

    @contents.setter
    def contents(self, value: Any) -> None:
        self._contents = parse_anything(value)

    @property
    def attributes(self) -> list[Attribute]:
        """The list of attributes affecting the tag.

        Each attribute is an instance of :class:`.Attribute`.
        """
        return self._attrs

    @property
    def wiki_markup(self) -> str | None:
        """The wikified version of a tag to show instead of HTML.

        If set to a value, this will be displayed instead of the brackets.
        For example, set to ``''`` to replace ``<i>`` or ``----`` to replace
        ``<hr>``.
        """
        return self._wiki_markup

    @wiki_markup.setter
    def wiki_markup(self, value: str | None) -> None:
        self._wiki_markup = str(value) if value else None
        if not value or not self.closing_wiki_markup:
            self._closing_wiki_markup = self._wiki_markup

    @property
    def self_closing(self) -> bool:
        """Whether the tag is self-closing with no content (like ``<br/>``)."""
        return self._self_closing

    @self_closing.setter
    def self_closing(self, value: bool) -> None:
        self._self_closing = bool(value)

    @property
    def invalid(self) -> bool:
        """Whether the tag starts with a backslash after the opening bracket.

        This makes the tag look like a lone close tag. It is technically
        invalid and is only parsable Wikicode when the tag itself is
        single-only, like ``<br>`` and ``<img>``. See
        :func:`.definitions.is_single_only`.
        """
        return self._invalid

    @invalid.setter
    def invalid(self, value: bool) -> None:
        self._invalid = bool(value)

    @property
    def implicit(self) -> bool:
        """Whether the tag is implicitly self-closing, with no ending slash.

        This is only possible for specific "single" tags like ``<br>`` and
        ``<li>``. See :func:`.definitions.is_single`. This field only has an
        effect if :attr:`self_closing` is also ``True``.
        """
        return self._implicit

    @implicit.setter
    def implicit(self, value: bool) -> None:
        self._implicit = bool(value)

    @property
    def padding(self) -> str:
        """Spacing to insert before the first closing ``>``."""
        return self._padding

    @padding.setter
    def padding(self, value: str | None) -> None:
        if not value:
            self._padding = ""
        else:
            value = str(value)
            if not value.isspace():
                raise ValueError("padding must be entirely whitespace")
            self._padding = value

    @property
    def closing_tag(self) -> Wikicode:
        """The closing tag, as a :class:`.Wikicode` object.

        This will usually equal :attr:`tag`, unless there is additional
        spacing, comments, or the like.
        """
        return self._closing_tag

    @closing_tag.setter
    def closing_tag(self, value: Any) -> None:
        self._closing_tag = parse_anything(value)

    @property
    def wiki_style_separator(self) -> str | None:
        """The separator between the padding and content in a wiki markup tag.

        Essentially the wiki equivalent of the TagCloseOpen.
        """
        return self._wiki_style_separator

    @wiki_style_separator.setter
    def wiki_style_separator(self, value: str | None) -> None:
        self._wiki_style_separator = str(value) if value else None

    @property
    def closing_wiki_markup(self) -> str | None:
        """The wikified version of the closing tag to show instead of HTML.

        If set to a value, this will be displayed instead of the close tag
        brackets. If the tag is :attr:`self_closing`, this is not displayed.
        If :attr:`wiki_markup` is set and this has not been set, this is set
        to the value of :attr:`wiki_markup`. If this has been set and
        :attr:`wiki_markup` is set to a ``False`` value, this is set to
        ``None``.
        """
        return self._closing_wiki_markup

    @closing_wiki_markup.setter
    def closing_wiki_markup(self, value: str | None) -> None:
        self._closing_wiki_markup = str(value) if value else None

    def has(self, name: str | Attribute | Wikicode) -> bool:
        """Return whether any attribute in the tag has the given *name*.

        Note that a tag may have multiple attributes with the same name, but
        only the last one is read by the MediaWiki parser.
        """
        for attr in self.attributes:
            if attr.name == name.strip():
                return True
        return False

    def get(self, name: str | Attribute | Wikicode) -> Attribute:
        """Get the attribute with the given *name*.

        The returned object is a :class:`.Attribute` instance. Raises
        :exc:`ValueError` if no attribute has this name. Since multiple
        attributes can have the same name, we'll return the last match, since
        all but the last are ignored by the MediaWiki parser.
        """
        for attr in reversed(self.attributes):
            if attr.name == name.strip():
                return attr
        raise ValueError(name)

    def add(
        self,
        name: Any,
        value: Any = None,
        quotes: str | None = '"',
        pad_first: str = " ",
        pad_before_eq: str = "",
        pad_after_eq: str = "",
    ) -> Attribute:
        """Add an attribute with the given *name* and *value*.

        *name* and *value* can be anything parsable by
        :func:`.utils.parse_anything`; *value* can be omitted if the attribute
        is valueless. If *quotes* is not ``None``, it should be a string
        (either ``"`` or ``'``) that *value* will be wrapped in (this is
        recommended). ``None`` is only legal if *value* contains no spacing.

        *pad_first*, *pad_before_eq*, and *pad_after_eq* are whitespace used as
        padding before the name, before the equal sign (or after the name if no
        value), and after the equal sign (ignored if no value), respectively.
        """
        if value is not None:
            value = parse_anything(value)
        quotes = Attribute.coerce_quotes(quotes)
        attr = Attribute(parse_anything(name), value, quotes)
        attr.pad_first = pad_first
        attr.pad_before_eq = pad_before_eq
        attr.pad_after_eq = pad_after_eq
        self.attributes.append(attr)
        return attr

    def remove(self, name: str) -> None:
        """Remove all attributes with the given *name*.

        Raises :exc:`ValueError` if none were found.
        """
        attrs = [attr for attr in self.attributes if attr.name == name.strip()]
        if not attrs:
            raise ValueError(name)
        for attr in attrs:
            self.attributes.remove(attr)



# Copyright (C) 2012-2025 Ben Kurtovic <ben.kurtovic@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
This module contains various "context" definitions, which are essentially flags
set during the tokenization process, either on the current parse stack (local
contexts) or affecting all stacks (global contexts). They represent the context
the tokenizer is in, such as inside a template's name definition, or inside a
level-two heading. This is used to determine what tokens are valid at the
current point and also if the current parsing route is invalid.

The tokenizer stores context as an integer, with these definitions bitwise OR'd
to set them, AND'd to check if they're set, and XOR'd to unset them. The
advantage of this is that contexts can have sub-contexts (as ``FOO == 0b11``
will cover ``BAR == 0b10`` and ``BAZ == 0b01``).

Local (stack-specific) contexts:

* :const:`TEMPLATE`

    * :const:`TEMPLATE_NAME`
    * :const:`TEMPLATE_PARAM_KEY`
    * :const:`TEMPLATE_PARAM_VALUE`

* :const:`ARGUMENT`

    * :const:`ARGUMENT_NAME`
    * :const:`ARGUMENT_DEFAULT`

* :const:`WIKILINK`

    * :const:`WIKILINK_TITLE`
    * :const:`WIKILINK_TEXT`

* :const:`EXT_LINK`

    * :const:`EXT_LINK_URI`
    * :const:`EXT_LINK_TITLE`

* :const:`HEADING`

    * :const:`HEADING_LEVEL_1`
    * :const:`HEADING_LEVEL_2`
    * :const:`HEADING_LEVEL_3`
    * :const:`HEADING_LEVEL_4`
    * :const:`HEADING_LEVEL_5`
    * :const:`HEADING_LEVEL_6`

* :const:`TAG`

    * :const:`TAG_OPEN`
    * :const:`TAG_ATTR`
    * :const:`TAG_BODY`
    * :const:`TAG_CLOSE`

* :const:`STYLE`

    * :const:`STYLE_ITALICS`
    * :const:`STYLE_BOLD`
    * :const:`STYLE_PASS_AGAIN`
    * :const:`STYLE_SECOND_PASS`

* :const:`DL_TERM`

* :const:`SAFETY_CHECK`

    * :const:`HAS_TEXT`
    * :const:`FAIL_ON_TEXT`
    * :const:`FAIL_NEXT`
    * :const:`FAIL_ON_LBRACE`
    * :const:`FAIL_ON_RBRACE`
    * :const:`FAIL_ON_EQUALS`
    * :const:`HAS_TEMPLATE`

* :const:`TABLE`

    * :const:`TABLE_OPEN`
    * :const:`TABLE_CELL_OPEN`
    * :const:`TABLE_CELL_STYLE`
    * :const:`TABLE_TD_LINE`
    * :const:`TABLE_TH_LINE`
    * :const:`TABLE_CELL_LINE_CONTEXTS`

* :const:`HTML_ENTITY`

Global contexts:

* :const:`GL_HEADING`

Aggregate contexts:

* :const:`FAIL`
* :const:`UNSAFE`
* :const:`DOUBLE`
* :const:`NO_WIKILINKS`
* :const:`NO_EXT_LINKS`

"""

from __future__ import annotations

# Local contexts:

TEMPLATE_NAME = 1 << 0
TEMPLATE_PARAM_KEY = 1 << 1
TEMPLATE_PARAM_VALUE = 1 << 2
TEMPLATE = TEMPLATE_NAME + TEMPLATE_PARAM_KEY + TEMPLATE_PARAM_VALUE

ARGUMENT_NAME = 1 << 3
ARGUMENT_DEFAULT = 1 << 4
ARGUMENT = ARGUMENT_NAME + ARGUMENT_DEFAULT

WIKILINK_TITLE = 1 << 5
WIKILINK_TEXT = 1 << 6
WIKILINK = WIKILINK_TITLE + WIKILINK_TEXT

EXT_LINK_URI = 1 << 7
EXT_LINK_TITLE = 1 << 8
EXT_LINK = EXT_LINK_URI + EXT_LINK_TITLE

HEADING_LEVEL_1 = 1 << 9
HEADING_LEVEL_2 = 1 << 10
HEADING_LEVEL_3 = 1 << 11
HEADING_LEVEL_4 = 1 << 12
HEADING_LEVEL_5 = 1 << 13
HEADING_LEVEL_6 = 1 << 14
HEADING = (
    HEADING_LEVEL_1
    + HEADING_LEVEL_2
    + HEADING_LEVEL_3
    + HEADING_LEVEL_4
    + HEADING_LEVEL_5
    + HEADING_LEVEL_6
)

TAG_OPEN = 1 << 15
TAG_ATTR = 1 << 16
TAG_BODY = 1 << 17
TAG_CLOSE = 1 << 18
TAG = TAG_OPEN + TAG_ATTR + TAG_BODY + TAG_CLOSE

STYLE_ITALICS = 1 << 19
STYLE_BOLD = 1 << 20
STYLE_PASS_AGAIN = 1 << 21
STYLE_SECOND_PASS = 1 << 22
STYLE = STYLE_ITALICS + STYLE_BOLD + STYLE_PASS_AGAIN + STYLE_SECOND_PASS

DL_TERM = 1 << 23

HAS_TEXT = 1 << 24
FAIL_ON_TEXT = 1 << 25
FAIL_NEXT = 1 << 26
FAIL_ON_LBRACE = 1 << 27
FAIL_ON_RBRACE = 1 << 28
FAIL_ON_EQUALS = 1 << 29
HAS_TEMPLATE = 1 << 30
SAFETY_CHECK = (
    HAS_TEXT
    + FAIL_ON_TEXT
    + FAIL_NEXT
    + FAIL_ON_LBRACE
    + FAIL_ON_RBRACE
    + FAIL_ON_EQUALS
    + HAS_TEMPLATE
)

TABLE_OPEN = 1 << 31
TABLE_CELL_OPEN = 1 << 32
TABLE_CELL_STYLE = 1 << 33
TABLE_ROW_OPEN = 1 << 34
TABLE_TD_LINE = 1 << 35
TABLE_TH_LINE = 1 << 36
TABLE_CELL_LINE_CONTEXTS = TABLE_TD_LINE + TABLE_TH_LINE + TABLE_CELL_STYLE
TABLE = (
    TABLE_OPEN
    + TABLE_CELL_OPEN
    + TABLE_CELL_STYLE
    + TABLE_ROW_OPEN
    + TABLE_TD_LINE
    + TABLE_TH_LINE
)

HTML_ENTITY = 1 << 37

# Global contexts:

GL_HEADING = 1 << 0

# Aggregate contexts:

FAIL = TEMPLATE + ARGUMENT + WIKILINK + EXT_LINK_TITLE + HEADING + TAG + STYLE + TABLE
UNSAFE = (
    TEMPLATE_NAME
    + WIKILINK_TITLE
    + EXT_LINK_TITLE
    + TEMPLATE_PARAM_KEY
    + ARGUMENT_NAME
    + TAG_CLOSE
)
DOUBLE = TEMPLATE_PARAM_KEY + TAG_CLOSE + TABLE_ROW_OPEN
NO_WIKILINKS = TEMPLATE_NAME + ARGUMENT_NAME + WIKILINK_TITLE + EXT_LINK_URI
NO_EXT_LINKS = TEMPLATE_NAME + ARGUMENT_NAME + WIKILINK_TITLE + EXT_LINK


def describe(context):
    """Return a string describing the given context value, for debugging."""
    flags = []
    for name, value in globals().items():
        if not isinstance(value, int) or name.startswith("GL_"):
            continue
        if bin(value).count("1") != 1:
            continue  # Hacky way to skip aggregate contexts
        if context & value:
            flags.append((name, value))
    flags.sort(key=lambda it: it[1])
    return "|".join(it[0] for it in flags)



# Copyright (C) 2012-2025 Ben Kurtovic <ben.kurtovic@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import re
from collections import defaultdict
from collections.abc import Generator, Mapping
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    TypeVar,
    overload,
)

from ..utils import parse_anything
from ._base import Node
from .extras import Parameter
from .html_entity import HTMLEntity
from .text import Text

if TYPE_CHECKING:
    from ..wikicode import Wikicode

__all__ = ["Template"]

# Used to allow None as a valid fallback value
_UNSET = object()

T = TypeVar("T")


class Template(Node):
    """Represents a template in wikicode, like ``{{foo}}``."""

    def __init__(self, name: Any, params: list[Parameter] | None = None):
        super().__init__()
        self.name = name
        self._params: list[Parameter] = params or []

    def __str__(self) -> str:
        if self.params:
            params = "|".join([str(param) for param in self.params])
            return "{{" + str(self.name) + "|" + params + "}}"
        return "{{" + str(self.name) + "}}"

    def __children__(self) -> Generator[Wikicode]:
        yield self.name
        for param in self.params:
            if param.showkey:
                yield param.name
            yield param.value

    def __strip__(self, **kwargs: Any) -> str | None:
        if kwargs.get("keep_template_params"):
            parts = [param.value.strip_code(**kwargs) for param in self.params]
            return " ".join(part for part in parts if part)
        return None

    def __showtree__(
        self,
        write: Callable[[str], None],
        get: Callable[[Wikicode], None],
        mark: Callable[[], None],
    ) -> None:
        write("{{")
        get(self.name)
        for param in self.params:
            write("    | ")
            mark()
            get(param.name)
            write("    = ")
            mark()
            get(param.value)
        write("}}")

    @staticmethod
    def _surface_escape(code: Wikicode, char: str) -> None:
        """Return *code* with *char* escaped as an HTML entity.

        The main use of this is to escape pipes (``|``) or equal signs (``=``)
        in parameter names or values so they are not mistaken for new
        parameters.
        """
        replacement = str(HTMLEntity(value=ord(char)))
        for node in code.filter_text(recursive=False):
            if char in node:
                code.replace(node, node.replace(char, replacement), False)

    @staticmethod
    def _select_theory(theories: dict[str, int]) -> str | None:
        """Return the most likely spacing convention given different options.

        Given a dictionary of convention options as keys and their occurrence
        as values, return the convention that occurs the most, or ``None`` if
        there is no clear preferred style.
        """
        if theories:
            values = tuple(theories.values())
            best = max(values)
            confidence = float(best) / sum(values)
            if confidence > 0.5:
                return tuple(theories.keys())[values.index(best)]
        return None

    @staticmethod
    def _blank_param_value(value: Wikicode) -> None:
        """Remove the content from *value* while keeping its whitespace.

        Replace *value*\\ 's nodes with two text nodes, the first containing
        whitespace from before its content and the second containing whitespace
        from after its content.
        """
        sval = str(value)
        if sval.isspace():
            before, after = "", sval
        else:
            match = re.search(r"^(\s*).*?(\s*)$", sval, re.DOTALL)
            assert match, sval
            before, after = match.group(1), match.group(2)
        value.nodes = [Text(before), Text(after)]

    def _get_spacing_conventions(
        self, use_names: bool
    ) -> tuple[str | None, str | None]:
        """Try to determine the whitespace conventions for parameters.

        This will examine the existing parameters and use
        :meth:`_select_theory` to determine if there are any preferred styles
        for how much whitespace to put before or after the value.
        """
        before_theories: defaultdict[str, int] = defaultdict(int)
        after_theories: defaultdict[str, int] = defaultdict(int)
        for param in self.params:
            if not param.showkey:
                continue
            if use_names:
                component = str(param.name)
            else:
                component = str(param.value)
            match = re.search(r"^(\s*).*?(\s*)$", component, re.DOTALL)
            assert match, component
            before, after = match.group(1), match.group(2)
            if not use_names and component.isspace() and "\n" in before:
                # If the value is empty, we expect newlines in the whitespace
                # to be after the content, not before it:
                before, after = before.split("\n", 1)
                after = "\n" + after
            before_theories[before] += 1
            after_theories[after] += 1

        before = self._select_theory(before_theories)
        after = self._select_theory(after_theories)
        return before, after

    def _fix_dependendent_params(self, i: int) -> None:
        """Unhide keys if necessary after removing the param at index *i*."""
        if not self.params[i].showkey:
            for param in self.params[i + 1 :]:
                if not param.showkey:
                    param.showkey = True

    def _remove_exact(self, needle: Parameter, keep_field: bool) -> None:
        """Remove a specific parameter, *needle*, from the template."""
        for i, param in enumerate(self.params):
            if param is needle:
                if keep_field:
                    self._blank_param_value(param.value)
                else:
                    self._fix_dependendent_params(i)
                    self.params.pop(i)
                return
        raise ValueError(needle)

    def _should_remove(self, i: int, name: str) -> bool:
        """Look ahead for a parameter with the same name, but hidden.

        If one exists, we should remove the given one rather than blanking it.
        """
        if self.params[i].showkey:
            following = self.params[i + 1 :]
            better_matches = [
                after.name.strip() == name and not after.showkey for after in following
            ]
            return any(better_matches)
        return False

    @property
    def name(self) -> Wikicode:
        """The name of the template, as a :class:`.Wikicode` object."""
        return self._name

    @name.setter
    def name(self, value: Any) -> None:
        self._name = parse_anything(value)

    @property
    def params(self) -> list[Parameter]:
        """The list of parameters contained within the template."""
        return self._params

    def has(self, name: str | Any, ignore_empty: bool = False) -> bool:
        """Return ``True`` if any parameter in the template is named *name*.

        With *ignore_empty*, ``False`` will be returned even if the template
        contains a parameter with the name *name*, if the parameter's value
        is empty. Note that a template may have multiple parameters with the
        same name, but only the last one is read by the MediaWiki parser.
        """
        name = str(name).strip()
        for param in self.params:
            if param.name.strip() == name:
                if ignore_empty and not param.value.strip():
                    continue
                return True
        return False

    def has_param(self, name: str | Any, ignore_empty: bool = False) -> bool:
        """Alias for :meth:`has`."""
        return self.has(name, ignore_empty)

    @overload
    def get(self, name: str | Any) -> Parameter: ...

    @overload
    def get(self, name: str | Any, default: T) -> Parameter | T: ...

    def get(self, name: str | Any, default: T = _UNSET) -> Parameter | T:
        """Get the parameter whose name is *name*.

        The returned object is a :class:`.Parameter` instance. Raises
        :exc:`ValueError` if no parameter has this name. If *default* is set,
        returns that instead. Since multiple parameters can have the same name,
        we'll return the last match, since the last parameter is the only one
        read by the MediaWiki parser.
        """
        name = str(name).strip()
        for param in reversed(self.params):
            if param.name.strip() == name:
                return param
        if default is _UNSET:
            raise ValueError(name)
        return default

    def __getitem__(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, name: str | Any
    ) -> Parameter:
        return self.get(name)

    def add(
        self,
        name: Any,
        value: Any,
        showkey: bool | None = None,
        before: Parameter | str | None = None,
        after: Parameter | str | None = None,
        preserve_spacing: bool = True,
    ) -> Parameter:
        """Add a parameter to the template with a given *name* and *value*.

        *name* and *value* can be anything parsable by
        :func:`.utils.parse_anything`; pipes and equal signs are automatically
        escaped from *value* when appropriate.

        If *name* is already a parameter in the template, we'll replace its
        value.

        If *showkey* is given, this will determine whether or not to show the
        parameter's name (e.g., ``{{foo|bar}}``'s parameter has a name of
        ``"1"`` but it is hidden); otherwise, we'll make a safe and intelligent
        guess.

        If *before* is given (either a :class:`.Parameter` object or a name),
        then we will place the parameter immediately before this one.
        Otherwise, it will be added at the end. If *before* is a name and
        exists multiple times in the template, we will place it before the last
        occurrence. If *before* is not in the template, :exc:`ValueError` is
        raised. The argument is ignored if *name* is an existing parameter.

        If *after* is given (either a :class:`.Parameter` object or a name),
        then we will place the parameter immediately after this one. If *after*
        is a name and exists multiple times in the template, we will place it
        after the last occurrence. If *after* is not in the template,
        :exc:`ValueError` is raised. The argument is ignored if *name* is an
        existing parameter or if a value is passed to *before*.

        If *preserve_spacing* is ``True``, we will try to preserve whitespace
        conventions around the parameter, whether it is new or we are updating
        an existing value. It is disabled for parameters with hidden keys,
        since MediaWiki doesn't strip whitespace in this case.
        """
        name, value = parse_anything(name), parse_anything(value)
        self._surface_escape(value, "|")

        if self.has(name):
            self.remove(name, keep_field=True)
            existing = self.get(name)
            if showkey is not None:
                existing.showkey = showkey
            if not existing.showkey:
                self._surface_escape(value, "=")
            nodes: list[Node | None] = list(existing.value.nodes)
            if preserve_spacing and existing.showkey:
                for i in range(2):  # Ignore empty text nodes
                    if not nodes[i]:
                        nodes[i] = None
                existing.value = parse_anything([nodes[0], value, nodes[1]])
            else:
                existing.value = value
            return existing

        if showkey is None:
            if Parameter.can_hide_key(name):
                int_name = int(str(name))
                int_keys = set()
                for param in self.params:
                    if not param.showkey:
                        int_keys.add(int(str(param.name)))
                expected = min(set(range(1, len(int_keys) + 2)) - int_keys)
                if expected == int_name:
                    showkey = False
                else:
                    showkey = True
            else:
                showkey = True
        if not showkey:
            self._surface_escape(value, "=")

        if preserve_spacing and showkey:
            before_n, after_n = self._get_spacing_conventions(use_names=True)
            before_v, after_v = self._get_spacing_conventions(use_names=False)
            name = parse_anything([before_n, name, after_n])
            value = parse_anything([before_v, value, after_v])

        param = Parameter(name, value, showkey)
        if before:
            assert after is None, "Cannot set a value for both 'before' and 'after'"
            if not isinstance(before, Parameter):
                before = self.get(before)
            self.params.insert(self.params.index(before), param)
        elif after:
            if not isinstance(after, Parameter):
                after = self.get(after)
            self.params.insert(self.params.index(after) + 1, param)
        else:
            self.params.append(param)
        return param

    def update(self, params: Mapping[Any, Any], **kwargs: Any) -> None:
        """Update the template with multiple parameters at once.

        - *params*: A dictionary mapping parameter names to values.
        - *kwargs*: Optional arguments that will be applied to all parameters, matching
          the same arguments in :meth:`add` (*showkey*, *before*, *after*,
          *preserve_spacing*).
        """
        for name, value in params.items():
            self.add(name, value, **kwargs)

    def __setitem__(self, name: Any, value: Any) -> Parameter:
        return self.add(name, value)

    def remove(self, param: Parameter | str | int, keep_field: bool = False) -> None:
        """Remove a parameter from the template, identified by *param*.

        If *param* is a :class:`.Parameter` object, it will be matched exactly,
        otherwise it will be treated like the *name* argument to :meth:`has`
        and :meth:`get`.

        If *keep_field* is ``True``, we will keep the parameter's name, but
        blank its value. Otherwise, we will remove the parameter completely.

        When removing a parameter with a hidden name, subsequent parameters
        with hidden names will be made visible. For example, removing ``bar``
        from ``{{foo|bar|baz}}`` produces ``{{foo|2=baz}}`` because
        ``{{foo|baz}}`` is incorrect.

        If the parameter shows up multiple times in the template and *param* is
        not a :class:`.Parameter` object, we will remove all instances of it
        (and keep only one if *keep_field* is ``True`` - either the one with a
        hidden name, if it exists, or the first instance).
        """
        if isinstance(param, Parameter):
            self._remove_exact(param, keep_field)
            return

        name = str(param).strip()
        removed = False
        to_remove = []

        for i, par in enumerate(self.params):
            if par.name.strip() == name:
                if keep_field:
                    if self._should_remove(i, name):
                        to_remove.append(i)
                    else:
                        self._blank_param_value(par.value)
                        keep_field = False
                else:
                    self._fix_dependendent_params(i)
                    to_remove.append(i)
                if not removed:
                    removed = True

        if not removed:
            raise ValueError(name)
        for i in reversed(to_remove):
            self.params.pop(i)

    def __delitem__(self, param: Parameter | str) -> None:
        return self.remove(param)



# Copyright (C) 2012-2025 Ben Kurtovic <ben.kurtovic@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
This package contains the actual wikicode parser, split up into two main
modules: the :mod:`.tokenizer` and the :mod:`.builder`. This module joins them
together into one interface.
"""

from __future__ import annotations

from .builder import Builder
from .errors import ParserError

try:
    from ._tokenizer import CTokenizer

    use_c = True
except ImportError:
    CTokenizer = None
    use_c = False

__all__ = ["use_c", "Parser", "ParserError"]


class Parser:
    """Represents a parser for wikicode.

    Actual parsing is a two-step process: first, the text is split up into a
    series of tokens by the :class:`.Tokenizer`, and then the tokens are
    converted into trees of :class:`.Wikicode` objects and :class:`.Node`\\ s
    by the :class:`.Builder`.

    Instances of this class or its dependents (:class:`.Tokenizer` and
    :class:`.Builder`) should not be shared between threads. :meth:`parse` can
    be called multiple times as long as it is not done concurrently. In
    general, there is no need to do this because parsing should be done through
    :func:`mwparserfromhell.parse`, which creates a new :class:`.Parser` object
    as necessary.
    """

    def __init__(self):
        if use_c and CTokenizer:
            self._tokenizer = CTokenizer()
        else:
            from .tokenizer import Tokenizer

            self._tokenizer = Tokenizer()
        self._builder = Builder()

    def parse(self, text, context=0, skip_style_tags=False):
        """Parse *text*, returning a :class:`.Wikicode` object tree.

        If given, *context* will be passed as a starting context to the parser.
        This is helpful when this function is used inside node attribute
        setters. For example, :class:`.ExternalLink`\\ 's
        :attr:`~.ExternalLink.url` setter sets *context* to
        :mod:`contexts.EXT_LINK_URI <.contexts>` to prevent the URL itself
        from becoming an :class:`.ExternalLink`.

        If *skip_style_tags* is ``True``, then ``''`` and ``'''`` will not be
        parsed, but instead will be treated as plain text.

        If there is an internal error while parsing, :exc:`.ParserError` will
        be raised.
        """
        tokens = self._tokenizer.tokenize(text, context, skip_style_tags)
        code = self._builder.build(tokens)
        return code



# Copyright (C) 2012-2025 Ben Kurtovic <ben.kurtovic@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
This module contains accessory functions for other parts of the library. Parser
users generally won't need stuff from here.
"""

from __future__ import annotations

__all__ = ["parse_anything"]

import typing
from typing import Any

if typing.TYPE_CHECKING:
    from .wikicode import Wikicode


def parse_anything(
    value: Any, context: int = 0, *, skip_style_tags: bool = False
) -> Wikicode:
    """Return a :class:`.Wikicode` for *value*, allowing multiple types.

    This differs from :meth:`.Parser.parse` in that we accept more than just a
    string to be parsed. Strings, bytes, integers (converted to strings),
    ``None``, existing :class:`.Node` or :class:`.Wikicode` objects, as well
    as an iterable of these types, are supported. This is used to parse input
    on-the-fly by various methods of :class:`.Wikicode` and others like
    :class:`.Template`, such as :meth:`wikicode.insert() <.Wikicode.insert>`
    or setting :meth:`template.name <.Template.name>`.

    Additional arguments are passed directly to :meth:`.Parser.parse`.
    """
    # pylint: disable=cyclic-import,import-outside-toplevel
    from .nodes import Node
    from .parser import Parser
    from .smart_list import SmartList
    from .wikicode import Wikicode

    if isinstance(value, Wikicode):
        return value
    if isinstance(value, Node):
        return Wikicode(SmartList([value]))
    if isinstance(value, str):
        return Parser().parse(value, context, skip_style_tags)
    if isinstance(value, bytes):
        return Parser().parse(value.decode("utf8"), context, skip_style_tags)
    if isinstance(value, int):
        return Parser().parse(str(value), context, skip_style_tags)
    if value is None:
        return Wikicode(SmartList())
    if hasattr(value, "read"):
        return parse_anything(value.read(), context, skip_style_tags=skip_style_tags)
    try:
        nodelist = SmartList()
        for item in value:
            nodelist += parse_anything(
                item, context, skip_style_tags=skip_style_tags
            ).nodes
        return Wikicode(nodelist)
    except TypeError as exc:
        error = (
            "Needs string, Node, Wikicode, file, int, None, or "
            "iterable of these, but got {0}: {1}"
        )
        raise ValueError(error.format(type(value).__name__, value)) from exc



# Copyright (C) 2012-2025 Ben Kurtovic <ben.kurtovic@gmail.com>
# Copyright (C) 2019-2020 Yuri Astrakhan <YuriAstrakhan@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import weakref
from typing import Any

from .utils import _SliceNormalizerMixIn, inheritdoc


class ListProxy(_SliceNormalizerMixIn, list):
    """Implement the ``list`` interface by getting elements from a parent.

    This is created by a :class:`.SmartList` object when slicing. It does not
    actually store the list at any time; instead, whenever the list is needed,
    it builds it dynamically using the :meth:`_render` method.
    """

    __slots__ = ("__weakref__", "_parent", "_sliceinfo")

    def __init__(self, parent, sliceinfo):
        super().__init__()
        self._parent = parent
        self._sliceinfo = sliceinfo

    def __reduce_ex__(self, protocol: Any) -> tuple:
        return (ListProxy, (self._parent, self._sliceinfo), ())

    def __setstate__(self, state: tuple) -> None:
        # Reregister with the parent
        child_ref = weakref.ref(self, self._parent._delete_child)
        self._parent._children[id(child_ref)] = (child_ref, self._sliceinfo)

    def __repr__(self):
        return repr(self._render())

    def __lt__(self, other):
        if isinstance(other, ListProxy):
            return self._render() < list(other)
        return self._render() < other

    def __le__(self, other):
        if isinstance(other, ListProxy):
            return self._render() <= list(other)
        return self._render() <= other

    def __eq__(self, other):
        if isinstance(other, ListProxy):
            return self._render() == list(other)
        return self._render() == other

    def __ne__(self, other):
        if isinstance(other, ListProxy):
            return self._render() != list(other)
        return self._render() != other

    def __gt__(self, other):
        if isinstance(other, ListProxy):
            return self._render() > list(other)
        return self._render() > other

    def __ge__(self, other):
        if isinstance(other, ListProxy):
            return self._render() >= list(other)
        return self._render() >= other

    def __bool__(self):
        return bool(self._render())

    def __len__(self):
        return max((self._stop - self._start) // self._step, 0)

    def __getitem__(self, key):
        if isinstance(key, slice):
            key = self._normalize_slice(key, clamp=True)
            keystart = min(self._start + key.start, self._stop)
            keystop = min(self._start + key.stop, self._stop)
            adjusted = slice(keystart, keystop, key.step)
            return self._parent[adjusted]
        return self._render()[key]

    def __setitem__(self, key, item):
        if isinstance(key, slice):
            key = self._normalize_slice(key, clamp=True)
            keystart = min(self._start + key.start, self._stop)
            keystop = min(self._start + key.stop, self._stop)
            adjusted = slice(keystart, keystop, key.step)
            self._parent[adjusted] = item
        else:
            length = len(self)
            if key < 0:
                key = length + key
            if key < 0 or key >= length:
                raise IndexError("list assignment index out of range")
            self._parent[self._start + key] = item

    def __delitem__(self, key):
        if isinstance(key, slice):
            key = self._normalize_slice(key, clamp=True)
            keystart = min(self._start + key.start, self._stop)
            keystop = min(self._start + key.stop, self._stop)
            adjusted = slice(keystart, keystop, key.step)
            del self._parent[adjusted]
        else:
            length = len(self)
            if key < 0:
                key = length + key
            if key < 0 or key >= length:
                raise IndexError("list assignment index out of range")
            del self._parent[self._start + key]

    def __iter__(self):
        i = self._start
        while i < self._stop:
            yield self._parent[i]
            i += self._step

    def __reversed__(self):
        i = self._stop - 1
        while i >= self._start:
            yield self._parent[i]
            i -= self._step

    def __contains__(self, item):
        return item in self._render()

    def __add__(self, other):
        return type(self._parent)(list(self) + other)

    def __radd__(self, other):
        return type(self._parent)(other + list(self))

    def __iadd__(self, other):
        self.extend(other)
        return self

    def __mul__(self, other):
        return type(self._parent)(list(self) * other)

    def __rmul__(self, other):
        return type(self._parent)(other * list(self))

    def __imul__(self, other):
        self.extend(list(self) * (other - 1))
        return self

    @property
    def _start(self):
        """The starting index of this list, inclusive."""
        return self._sliceinfo[0]

    @property
    def _stop(self):
        """The ending index of this list, exclusive."""
        if self._sliceinfo[1] is None:
            return len(self._parent)
        return self._sliceinfo[1]

    @property
    def _step(self):
        """The number to increase the index by between items."""
        return self._sliceinfo[2]

    def _render(self):
        """Return the actual list from the stored start/stop/step."""
        return list(self._parent)[self._start : self._stop : self._step]

    @inheritdoc
    def append(self, item):
        self._parent.insert(self._stop, item)

    @inheritdoc
    def count(self, item):
        return self._render().count(item)

    @inheritdoc
    def index(self, item, start=None, stop=None):
        if start is not None:
            if stop is not None:
                return self._render().index(item, start, stop)
            return self._render().index(item, start)
        return self._render().index(item)

    @inheritdoc
    def extend(self, item):
        self._parent[self._stop : self._stop] = item

    @inheritdoc
    def insert(self, index, item):
        if index < 0:
            index = len(self) + index
        self._parent.insert(self._start + index, item)

    @inheritdoc
    def pop(self, index=None):
        length = len(self)
        if index is None:
            index = length - 1
        elif index < 0:
            index = length + index
        if index < 0 or index >= length:
            raise IndexError("pop index out of range")
        return self._parent.pop(self._start + index)

    @inheritdoc
    def remove(self, item):
        index = self.index(item)
        del self._parent[self._start + index]

    @inheritdoc
    def reverse(self):
        item = self._render()
        item.reverse()
        self._parent[self._start : self._stop : self._step] = item

    @inheritdoc
    def sort(self, *, key=None, reverse=None):
        item = self._render()
        kwargs = {}
        if key is not None:
            kwargs["key"] = key
        if reverse is not None:
            kwargs["reverse"] = reverse
        item.sort(**kwargs)
        self._parent[self._start : self._stop : self._step] = item



# Copyright (C) 2012-2025 Ben Kurtovic <ben.kurtovic@gmail.com>
# Copyright (C) 2019-2020 Yuri Astrakhan <YuriAstrakhan@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from typing import Any
from weakref import ref

from .list_proxy import ListProxy
from .utils import _SliceNormalizerMixIn, inheritdoc


class SmartList(list, _SliceNormalizerMixIn):
    """Implements the ``list`` interface with special handling of sublists.

    When a sublist is created (by ``list[i:j]``), any changes made to this
    list (such as the addition, removal, or replacement of elements) will be
    reflected in the sublist, or vice-versa, to the greatest degree possible.
    This is implemented by having sublists - instances of the
    :class:`.ListProxy` type - dynamically determine their elements by storing
    their slice info and retrieving that slice from the parent. Methods that
    change the size of the list also change the slice info. For example::

        >>> parent = SmartList([0, 1, 2, 3])
        >>> parent
        [0, 1, 2, 3]
        >>> child = parent[2:]
        >>> child
        [2, 3]
        >>> child.append(4)
        >>> child
        [2, 3, 4]
        >>> parent
        [0, 1, 2, 3, 4]
    """

    __slots__ = ("_children",)

    def __new__(cls, *args, **kwargs):
        obj = super().__new__(cls, *args, **kwargs)
        obj._children = {}
        return obj

    def __reduce_ex__(self, protocol: Any) -> tuple:
        # Detach children when pickling
        return (SmartList, (), None, iter(self))

    def __getitem__(self, key):
        if not isinstance(key, slice):
            return super().__getitem__(key)
        key = self._normalize_slice(key, clamp=False)
        sliceinfo = [key.start, key.stop, key.step]
        child = ListProxy(self, sliceinfo)
        child_ref = ref(child, self._delete_child)
        self._children[id(child_ref)] = (child_ref, sliceinfo)
        return child

    def __setitem__(self, key, item):
        if not isinstance(key, slice):
            super().__setitem__(key, item)
            return
        item = list(item)
        super().__setitem__(key, item)
        key = self._normalize_slice(key, clamp=True)
        diff = len(item) + (key.start - key.stop) // key.step
        if not diff:
            return
        for child, (start, stop, _step) in self._children.values():
            if start > key.stop:
                self._children[id(child)][1][0] += diff
            if stop is not None and stop >= key.stop:
                self._children[id(child)][1][1] += diff

    def __delitem__(self, key):
        super().__delitem__(key)
        if isinstance(key, slice):
            key = self._normalize_slice(key, clamp=True)
        else:
            key = slice(key, key + 1, 1)
        diff = (key.stop - key.start) // key.step
        for child, (start, stop, _step) in self._children.values():
            if start > key.start:
                self._children[id(child)][1][0] -= diff
            if stop is not None and stop >= key.stop:
                self._children[id(child)][1][1] -= diff

    def __add__(self, other):
        return SmartList(list(self) + other)

    def __radd__(self, other):
        return SmartList(other + list(self))

    def __iadd__(self, other):
        self.extend(other)
        return self

    def _delete_child(self, child_ref):
        """Remove a child reference that is about to be garbage-collected."""
        del self._children[id(child_ref)]

    def _detach_children(self):
        """Remove all children and give them independent parent copies."""
        children = [val[0] for val in self._children.values()]
        for child in children:
            child()._parent = list(self)
        self._children.clear()

    @inheritdoc
    def append(self, item):
        head = len(self)
        self[head:head] = [item]

    @inheritdoc
    def extend(self, item):
        head = len(self)
        self[head:head] = item

    @inheritdoc
    def insert(self, index, item):
        self[index:index] = [item]

    @inheritdoc
    def pop(self, index=None):
        if index is None:
            index = len(self) - 1
        item = self[index]
        del self[index]
        return item

    @inheritdoc
    def remove(self, item):
        del self[self.index(item)]

    @inheritdoc
    def reverse(self):
        self._detach_children()
        super().reverse()

    @inheritdoc
    def sort(self, *, key=None, reverse=None):
        self._detach_children()
        kwargs = {}
        if key is not None:
            kwargs["key"] = key
        if reverse is not None:
            kwargs["reverse"] = reverse
        super().sort(**kwargs)



# Copyright (C) 2012-2025 Ben Kurtovic <ben.kurtovic@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Generator
from typing import TYPE_CHECKING, Any, Callable

from ..utils import parse_anything
from ._base import Node

if TYPE_CHECKING:
    from ..wikicode import Wikicode


class Heading(Node):
    """Represents a section heading in wikicode, like ``== Foo ==``."""

    def __init__(self, title: Any, level: int):
        super().__init__()
        self.title = title  # pyright: ignore[reportIncompatibleMethodOverride]
        self.level = level

    def __str__(self) -> str:
        return ("=" * self.level) + str(self.title) + ("=" * self.level)

    def __children__(self) -> Generator[Wikicode, None, None]:
        yield self.title

    def __strip__(self, **kwargs: Any) -> str | None:
        return self.title.strip_code(**kwargs)

    def __showtree__(
        self,
        write: Callable[[str], None],
        get: Callable[[Wikicode], None],
        mark: Callable[[], None],
    ) -> None:
        write("=" * self.level)
        get(self.title)
        write("=" * self.level)

    @property
    def title(self) -> Wikicode:
        """The title of the heading, as a :class:`.Wikicode` object."""
        return self._title

    @title.setter
    def title(self, value: Any) -> None:
        self._title = parse_anything(value)

    @property
    def level(self) -> int:
        """The heading level, as an integer between 1 and 6, inclusive."""
        return self._level

    @level.setter
    def level(self, value: int) -> None:
        value = int(value)
        if value < 1 or value > 6:
            raise ValueError(value)
        self._level = value



# Copyright (C) 2012-2025 Ben Kurtovic <ben.kurtovic@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
This module contains the token definitions that are used as an intermediate
parsing data type - they are stored in a flat list, with each token being
identified by its type and optional attributes. The token list is generated in
a syntactically valid form by the :class:`.Tokenizer`, and then converted into
the :class`.Wikicode` tree by the :class:`.Builder`.
"""

from __future__ import annotations

__all__ = ["Token"]


class Token(dict):
    """A token stores the semantic meaning of a unit of wikicode."""

    def __repr__(self):
        args = []
        for key, value in self.items():
            if isinstance(value, str) and len(value) > 100:
                args.append(key + "=" + repr(value[:97] + "..."))
            else:
                args.append(key + "=" + repr(value))
        return "{}({})".format(type(self).__name__, ", ".join(args))

    def __eq__(self, other):
        return isinstance(other, type(self)) and dict.__eq__(self, other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __getattr__(self, key):
        return self.get(key)

    def __setattr__(self, key, value):
        self[key] = value

    def __delattr__(self, key):
        del self[key]


def make(name):
    """Create a new Token class using ``type()`` and add it to ``__all__``."""
    __all__.append(name)
    return type(name, (Token,), {})


Text = make("Text")

TemplateOpen = make("TemplateOpen")  # {{
TemplateParamSeparator = make("TemplateParamSeparator")  # |
TemplateParamEquals = make("TemplateParamEquals")  # =
TemplateClose = make("TemplateClose")  # }}

ArgumentOpen = make("ArgumentOpen")  # {{{
ArgumentSeparator = make("ArgumentSeparator")  # |
ArgumentClose = make("ArgumentClose")  # }}}

WikilinkOpen = make("WikilinkOpen")  # [[
WikilinkSeparator = make("WikilinkSeparator")  # |
WikilinkClose = make("WikilinkClose")  # ]]

ExternalLinkOpen = make("ExternalLinkOpen")  # [
ExternalLinkSeparator = make("ExternalLinkSeparator")  #
ExternalLinkClose = make("ExternalLinkClose")  # ]

HTMLEntityStart = make("HTMLEntityStart")  # &
HTMLEntityNumeric = make("HTMLEntityNumeric")  # #
HTMLEntityHex = make("HTMLEntityHex")  # x
HTMLEntityEnd = make("HTMLEntityEnd")  # ;

HeadingStart = make("HeadingStart")  # =...
HeadingEnd = make("HeadingEnd")  # =...

CommentStart = make("CommentStart")  # <!--
CommentEnd = make("CommentEnd")  # -->

TagOpenOpen = make("TagOpenOpen")  # <
TagAttrStart = make("TagAttrStart")
TagAttrEquals = make("TagAttrEquals")  # =
TagAttrQuote = make("TagAttrQuote")  # ", '
TagCloseOpen = make("TagCloseOpen")  # >
TagCloseSelfclose = make("TagCloseSelfclose")  # />
TagOpenClose = make("TagOpenClose")  # </
TagCloseClose = make("TagCloseClose")  # >

del make



# Copyright (C) 2012-2025 Ben Kurtovic <ben.kurtovic@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from ._base import Node

if TYPE_CHECKING:
    from ..wikicode import Wikicode

__all__ = ["Comment"]


class Comment(Node):
    """Represents a hidden HTML comment, like ``<!-- foobar -->``."""

    def __init__(self, contents: Wikicode | str):
        super().__init__()
        self.contents = contents

    def __str__(self) -> str:
        return "<!--" + self.contents + "-->"

    @property
    def contents(self) -> str:
        """The hidden text contained between ``<!--`` and ``-->``."""
        return self._contents

    @contents.setter
    def contents(self, value: Any) -> None:
        self._contents = str(value)



# Copyright (C) 2012-2025 Ben Kurtovic <ben.kurtovic@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from collections.abc import Generator
from typing import TYPE_CHECKING, Any, Callable

from ..utils import parse_anything
from ._base import Node

if TYPE_CHECKING:
    from ..wikicode import Wikicode

__all__ = ["ExternalLink"]


class ExternalLink(Node):
    """Represents an external link, like ``[http://example.com/ Example]``."""

    def __init__(
        self,
        url: Any,
        title: Any = None,
        brackets: bool = True,
        suppress_space: bool = False,
    ):
        super().__init__()
        self.url = url
        self.title = title  # pyright: ignore[reportIncompatibleMethodOverride]
        self.brackets = brackets
        self.suppress_space = suppress_space

    def __str__(self) -> str:
        if self.brackets:
            if self.title is not None:
                if self.suppress_space is True:
                    return "[" + str(self.url) + str(self.title) + "]"
                return "[" + str(self.url) + " " + str(self.title) + "]"
            return "[" + str(self.url) + "]"
        return str(self.url)

    def __children__(self) -> Generator[Wikicode, None, None]:
        yield self.url
        if self.title is not None:
            yield self.title

    def __strip__(self, **kwargs: Any) -> str | None:
        if self.brackets:
            if self.title:
                return self.title.strip_code(**kwargs)
            return None
        return self.url.strip_code(**kwargs)

    def __showtree__(
        self,
        write: Callable[[str], None],
        get: Callable[[Wikicode], None],
        mark: Callable[[], None],
    ) -> None:
        if self.brackets:
            write("[")
        get(self.url)
        if self.title is not None:
            get(self.title)
        if self.brackets:
            write("]")

    @property
    def url(self) -> Wikicode:
        """The URL of the link target, as a :class:`.Wikicode` object."""
        return self._url

    @url.setter
    def url(self, value: Any) -> None:
        # pylint: disable=import-outside-toplevel
        from ..parser import contexts

        self._url = parse_anything(value, contexts.EXT_LINK_URI)

    @property
    def title(self) -> Wikicode | None:
        """The link title (if given), as a :class:`.Wikicode` object."""
        return self._title

    @title.setter
    def title(self, value: Any) -> None:
        self._title = None if value is None else parse_anything(value)

    @property
    def brackets(self) -> bool:
        """Whether to enclose the URL in brackets or display it straight."""
        return self._brackets

    @brackets.setter
    def brackets(self, value: bool) -> None:
        self._brackets = bool(value)



# Copyright (C) 2012-2025 Ben Kurtovic <ben.kurtovic@gmail.com>
# Copyright (C) 2019-2020 Yuri Astrakhan <YuriAstrakhan@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import sys
from collections.abc import Sized
from typing import Literal, overload

__all__ = []


def inheritdoc(method):
    """Set __doc__ of *method* to __doc__ of *method* in its parent class.

    Since this is used on :class:`.SmartList`, the "parent class" used is
    ``list``. This function can be used as a decorator.
    """
    method.__doc__ = getattr(list, method.__name__).__doc__
    return method


class _SliceNormalizerMixIn(Sized):
    """MixIn that provides a private method to normalize slices."""

    __slots__ = ()

    @overload
    def _normalize_slice(
        self,
        key: slice[int | None, int | None, int | None],
        clamp: Literal[False] = False,
    ) -> slice[int, int | None, int]: ...

    @overload
    def _normalize_slice(
        self, key: slice[int | None, int | None, int | None], clamp: Literal[True]
    ) -> slice[int, int, int]: ...

    def _normalize_slice(
        self, key: slice[int | None, int | None, int | None], clamp: bool = False
    ) -> slice[int, int | None, int]:
        """Return a slice equivalent to the input *key*, standardized."""
        if key.start is None:
            start = 0
        else:
            start = (len(self) + key.start) if key.start < 0 else key.start
        if key.stop is None or key.stop == sys.maxsize:
            stop = len(self) if clamp else None
        else:
            stop = (len(self) + key.stop) if key.stop < 0 else key.stop
        return slice(start, stop, key.step or 1)



# Copyright (C) 2012-2025 Ben Kurtovic <ben.kurtovic@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
This package contains :class:`.Wikicode` "nodes", which represent a single unit
of wikitext, such as a Template, an HTML tag, a Heading, or plain text. The
node "tree" is far from flat, as most types can contain additional
:class:`.Wikicode` types within them - and with that, more nodes. For example,
the name of a :class:`.Template` is a :class:`.Wikicode` object that can
contain text or more templates.
"""

from __future__ import annotations

from . import extras
from ._base import Node
from .argument import Argument
from .comment import Comment
from .external_link import ExternalLink
from .heading import Heading
from .html_entity import HTMLEntity
from .tag import Tag
from .template import Template
from .text import Text
from .wikilink import Wikilink

__all__ = [
    "Argument",
    "Comment",
    "ExternalLink",
    "HTMLEntity",
    "Heading",
    "Node",
    "Tag",
    "Template",
    "Text",
    "Wikilink",
    "extras",
]



# Copyright (C) 2012-2025 Ben Kurtovic <ben.kurtovic@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import html.entities as htmlentities
from typing import Any

from ._base import Node

__all__ = ["HTMLEntity"]


class HTMLEntity(Node):
    """Represents an HTML entity, like ``&nbsp;``, either named or unnamed."""

    def __init__(
        self,
        value: Any,
        named: bool | None = None,
        hexadecimal: bool = False,
        hex_char: str = "x",
    ):
        super().__init__()
        self._value = str(value)
        if named is None:  # Try to guess whether or not the entity is named
            try:
                int(value)
                self._named = False
                self._hexadecimal = False
            except ValueError:
                try:
                    int(value, 16)
                    self._named = False
                    self._hexadecimal = True
                except ValueError:
                    self._named = True
                    self._hexadecimal = False
        else:
            self._named = named
            self._hexadecimal = hexadecimal
        self._hex_char = hex_char

    def __str__(self) -> str:
        if self.named:
            return f"&{self.value};"
        if self.hexadecimal:
            return f"&#{self.hex_char}{self.value};"
        return f"&#{self.value};"

    def __strip__(self, **kwargs: Any) -> str | None:
        if kwargs.get("normalize"):
            return self.normalize()
        return str(self)

    @property
    def value(self) -> str:
        """The string value of the HTML entity."""
        return self._value

    @value.setter
    def value(self, newval: Any) -> None:
        newval = str(newval)
        try:
            int(newval)
        except ValueError:
            try:
                intval = int(newval, 16)
            except ValueError:
                if newval not in htmlentities.entitydefs:
                    raise ValueError(
                        f"entity value {newval!r} is not a valid name"
                    ) from None
                self._named = True
                self._hexadecimal = False
            else:
                if intval < 0 or intval > 0x10FFFF:
                    raise ValueError(
                        f"entity value 0x{intval:x} is not in range(0x110000)"
                    ) from None
                self._named = False
                self._hexadecimal = True
        else:
            test = int(newval, 16 if self.hexadecimal else 10)
            if test < 0 or test > 0x10FFFF:
                raise ValueError(f"entity value {test} is not in range(0x110000)")
            self._named = False
        self._value = newval

    @property
    def named(self) -> bool:
        """Whether the entity is a string name for a codepoint or an integer.

        For example, ``&Sigma;``, ``&#931;``, and ``&#x3a3;`` refer to the same
        character, but only the first is "named", while the others are integer
        representations of the codepoint.
        """
        return self._named

    @named.setter
    def named(self, newval: bool) -> None:
        newval = bool(newval)
        if newval and self.value not in htmlentities.entitydefs:
            raise ValueError(f"entity value {self.value!r} is not a valid name")

        if not newval:
            try:
                int(self.value, 16)
            except ValueError as exc:
                raise ValueError(
                    f"current entity value {self.value!r} is not a valid Unicode codepoint"
                ) from exc
        self._named = newval

    @property
    def hexadecimal(self) -> bool:
        """If unnamed, this is whether the value is hexadecimal or decimal."""
        return self._hexadecimal

    @hexadecimal.setter
    def hexadecimal(self, newval: bool) -> None:
        newval = bool(newval)
        if newval and self.named:
            raise ValueError("a named entity cannot be hexadecimal")
        self._hexadecimal = newval

    @property
    def hex_char(self) -> str:
        """If the value is hexadecimal, this is the letter denoting that.

        For example, the hex_char of ``"&#x1234;"`` is ``"x"``, whereas the
        hex_char of ``"&#X1234;"`` is ``"X"``. Lowercase and uppercase ``x``
        are the only values supported.
        """
        return self._hex_char

    @hex_char.setter
    def hex_char(self, newval: str) -> None:
        newval = str(newval)
        if newval not in ("x", "X"):
            raise ValueError(newval)
        self._hex_char = newval

    def normalize(self) -> str:
        """Return the unicode character represented by the HTML entity."""
        if self.named:
            return chr(htmlentities.name2codepoint[self.value])
        if self.hexadecimal:
            return chr(int(self.value, 16))
        return chr(int(self.value))
